use clap::Parser;
use easymacros::x11_safe_wrapper::{Keysym, XDisplay};
use std::ffi::CString;
use std::process::{Command, exit};
use std::time::Duration;
use std::{fs, thread};
use std::io::stdin;

/// Macro player module for easymacros. It's partially compatible with xmacro macros, with aim for full compatibility.
#[derive(Parser, Debug)]
#[clap(author, version, about, long_about = None)]
struct Args {
    /// The file that contains the macro to run.
    #[clap(value_parser, value_name = "input_file", value_hint = clap::ValueHint::FilePath)]
    input_file: Option<std::path::PathBuf>,
    /// Display to run the macro on. This uses the $DISPLAY environment variable by default.
    #[clap(short, long)]
    display: Option<String>,
}

fn main() {
    let args = Args::parse();

    let display = get_remote(args.display);

    if let Some(input_file_path) = args.input_file {
        let input_file_contents = fs::read_to_string(input_file_path)
            .expect("Couldn't read macro file");

        for instruction in input_file_contents.lines() {
            run_instruction(instruction, &display);
        }
    } else {
        let stdin = stdin();

        loop {
            let mut line = String::new();
            stdin.read_line(&mut line).expect("Couldn't read line from stdin");
            // Without this it crashes because apparently it doesn't properly read the next input line?
            println!();
            line = line.trim().to_string();
            run_instruction(&*line, &display);
        }
    }



    display.close();
}

fn get_remote(display_name: Option<String>) -> XDisplay {
    let display = XDisplay::open(display_name);

    if !display.has_xtest() {
        eprintln!("XTest not supported!");
        display.close();
        exit(1)
    }

    display.grab_control();
    display.sync();

    display
}

fn run_instruction(instruction: &str, display: &XDisplay) {
    let instruction: Vec<&str> = instruction.split(' ').collect();

    match instruction[0] {
        "Delay" => thread::sleep(Duration::from_millis(instruction[1].parse().unwrap())),
        "ButtonPress" => display.send_fake_buttonpress(instruction[1].parse().unwrap()),
        "ButtonRelease" => display.send_fake_buttonrelease(instruction[1].parse().unwrap()),
        "MotionNotify" => display
            .send_fake_motion_event(instruction[1].parse().unwrap(), instruction[2].parse().unwrap()),
        "KeyCodePress" => display.send_fake_keypress_from_code(instruction[1].parse().unwrap()),
        "KeyCodeRelease" => display.send_fake_keyrelease_from_code(instruction[1].parse().unwrap()),
        "KeySymPress" => display.send_fake_keypress_from_keysym(instruction[1].parse().unwrap()),
        "KeySymRelease" => {
            display.send_fake_keyrelease_from_keysym(instruction[1].parse().unwrap())
        }
        "KeySym" => {
            let key: Keysym = instruction[1].parse().unwrap();
            display.send_fake_keypress_from_keysym(key);
            display.send_fake_keyrelease_from_keysym(key);
        }
        "KeyStrPress" => {
            display.send_fake_keypress_from_string(CString::new(instruction[1]).unwrap().as_bytes())
        }
        "KeyStrRelease" => display
            .send_fake_keyrelease_from_string(CString::new(instruction[1]).unwrap().as_bytes()),
        "KeyStr" => {
            let keystring = CString::new(instruction[1]).unwrap();
            display.send_fake_keypress_from_string(keystring.as_bytes());
            display.send_fake_keyrelease_from_string(keystring.as_bytes());
        }
        "String" => {
            println!("Strings are currently not supported.");
        }
        "ExecBlock" | "ExecNoBlock" => {
            let mut command = Command::new(instruction[1]);
            for arg in &instruction[2..] {
                command.arg(arg);
            }
            if instruction[0] == "ExecBlock" {
                command.status();
            } else {
                command.spawn();
            }
        }
        c => {
            panic!("Unknown command {:?}", instruction)
        }
    }
}