extern crate core;

use std::os::raw::{c_char, c_uchar, c_uint};
use std::process::{exit, ExitCode};
use std::ptr::{addr_of, slice_from_raw_parts};
use std::{io, slice, thread};
use std::io::Write;
use std::ffi::c_void;
use std::fmt::{format};
use std::fs::{File, OpenOptions};
use clap::Parser;
use x11::keysym::XK_Escape;
use x11::xinput2::XIGrabModeSync;
use x11::xlib::{ButtonPress, ButtonRelease, CurrentTime, GrabModeAsync, GrabModeSync, GrabSuccess, KeyCode, KeyPress, KeyPressMask, KeyRelease, MotionNotify, SyncPointer, XEvent, XFree, XKeyEvent, XKeyPressedEvent, XPointer};
use x11::xrecord::{XRecordAllocRange, XRecordContext, XRecordCreateContext, XRecordDisableContext, XRecordEndOfData, XRecordFreeData, XRecordInterceptData, XRecordStartOfData};
use easymacros::x11_safe_wrapper::{Keycode, XDisplay};

/// Macro recording module for easymacros. Outputs are partially compatible with xmacro.
#[derive(Parser, Debug)]
#[clap(author, version, about, long_about = None)]
struct Args {
	/// The file to record the macro to. Defaults to writing to stdout.
	#[clap(value_parser, value_name = "output_file", value_hint = clap::ValueHint::FilePath)]
	output_file: Option<std::path::PathBuf>,
	/// Display to run the macro on. This uses the $DISPLAY environment variable by default.
	#[clap(short, long)]
	display: Option<String>,
}
fn main() {
	let args = Args::parse();

	let display = XDisplay::open(args.display.clone());
	let recorded_display = XDisplay::open(args.display.clone());

	let stop_key = get_stop_key(display);

	let screen = display.get_default_screen();
	dbg!(stop_key);

	let mut outfile: Box<dyn Write> = if let Some(outfile) = args.output_file {
		Box::new(File::create(outfile).expect("Failed to create output file"))
	} else {
		Box::new(io::stdout())
	};

	ev_loop(display, recorded_display, screen, stop_key, outfile);
	display.close();
}

fn get_stop_key(display: XDisplay) -> Keycode {
	let screen = display.get_default_screen();

	let root = display.get_root_window(screen);
	let potential_err = display.grab_keyboard(root, false, GrabModeSync, GrabModeAsync, CurrentTime);

	if potential_err != GrabSuccess {
		eprintln!("Couldn't grab keyboard!");
		exit(1);
	}

	println!("Press the key you want to use to stop recording the macro.");

	let mut stop_key: Keycode = XK_Escape;

	loop {
		display.allow_events(SyncPointer, CurrentTime);
		let ev = display.window_event(root, KeyPressMask);

		unsafe {
			match ev {
				XEvent { key } => {
					stop_key = key.keycode;
					break;
				}
				_ => {},
			}
		}
	}

	display.ungrab_keyboard(CurrentTime);
	display.ungrab_pointer(CurrentTime);

	stop_key
}

fn ev_loop(display: XDisplay, recordeddpy: XDisplay, screen: i32, stop_key: Keycode, mut outfile: Box<dyn Write>) {
	let root = display.get_root_window(screen);
	let protocol_ranges = unsafe { XRecordAllocRange() };

	let pointer_pos = display.query_pointer_pos();

	if pointer_pos != (-1, -1) {
		writeln!(&mut outfile, "MotionNotify {} {}", pointer_pos.0, pointer_pos.1)
			.expect("Failed to write to outfile");
	}

	let ctx = recordeddpy.create_record_context(protocol_ranges);
	let ev_cb_data = EvCallbackData {
		outfile,
		xdpy: display,
		recdpy: recordeddpy,
		ctx,
		stop_key,
		ev_nr: 0,
		working: true,
		x: pointer_pos.0 as i16,
		y: pointer_pos.1 as i16,
		no_keypress_yet: true,
		last_event: 0,
		moving: false,
	};

	if !recordeddpy.enable_context_async(ctx, Some(ev_callback), addr_of!(ev_cb_data) as *mut c_char) {
		panic!("Failed to enable record context")
	}
	while ev_cb_data.working {
		recordeddpy.process_replies();
	}

	display.disable_context(ctx);
	display.free_context(ctx);
	unsafe { XFree(protocol_ranges as *mut c_void) };
}

#[repr(C)]
pub struct EvCallbackData {
	outfile: Box<dyn Write>,
	xdpy: XDisplay,
	recdpy: XDisplay,
	stop_key: Keycode,
	ev_nr: u32,
	working: bool,
	ctx: XRecordContext,
	x: i16,
	y: i16,
	no_keypress_yet: bool,
	last_event: u64,
	moving: bool,
}


unsafe extern "C" fn ev_callback(closure: *mut c_char, intercept_data: *mut XRecordInterceptData) {
	let data = &mut *(closure as *mut EvCallbackData);
	let intercept_data = &mut *intercept_data;

	if intercept_data.category == XRecordStartOfData {
		println!("Got start of data!");
		XRecordFreeData(intercept_data);
		return;
	} else if intercept_data.category == XRecordEndOfData {
		println!("Got end of data!");
		XRecordFreeData(intercept_data);
		return;
	}

	let ev_type = *(intercept_data.data as *const u8);
	if data.x == -1 || data.y == -1 {
		if ev_type == MOTIONNOTIFY_U8 {
			data.x = *((intercept_data.data as usize + std::mem::size_of::<i16>() * 10) as *const i16);
			data.y = *((intercept_data.data as usize + std::mem::size_of::<i16>() * 11) as *const i16);
			println!("MotionNotify {} {}", data.x, data.y);
		} else {
			println!("Move your cursor so the macro can start with a fixed cursor position!");
			println!("Skipping event...");
		}
	} else if data.no_keypress_yet && ev_type == KEYRELEASE_U8 {
		println!("Skipping KeyRelease without recorded KeyPress...");
	} else {
		// if
		match ev_type {
			MOTIONNOTIFY_U8 => {
				let root_x = *((intercept_data.data as usize + std::mem::size_of::<i16>() * 10) as *const i16);
				let root_y = *((intercept_data.data as usize + std::mem::size_of::<i16>() * 11) as *const i16);
				data.x = root_x;
				data.y = root_y;

				if !data.moving {
					data.moving = true;
				}
			}
			KEYPRESS_U8 | KEYRELEASE_U8 => {
				let kc: u8 = *((intercept_data.data as usize + 1) as *const u8);
				let keyname = data.xdpy.keycode_to_string(kc as u32);

				if ev_type == KEYPRESS_U8 && kc == data.stop_key as u8 {
					// println!("stop key detected, recording done!");
					data.working = false;
				} else {
					if ev_type == KEYPRESS_U8 {
						data.no_keypress_yet = false;
					}

					if (intercept_data.server_time - data.last_event) != 0 {
						println!("test");
						writeln!(&mut data.outfile,"Delay {}", intercept_data.server_time - data.last_event)
							.expect("Failed to write to outfile.");
						data.last_event = intercept_data.server_time;
					}

					if data.moving {
						writeln!(&mut data.outfile, "MotionNotify {} {}", data.x, data.y)
							.expect("Failed to write to outfile.");
						data.moving = false;
					}

					writeln!(
						&mut data.outfile,
						"{} {}",
						if ev_type == KEYPRESS_U8 { "KeyStrPress" } else { "KeyRelease" },
						keyname.to_str().unwrap()
					).expect("Failed to write to outfile.");
				}
			}
			BUTTONPRESS_U8 | BUTTONRELEASE_U8 => {
				let bc: u8 = *((intercept_data.data as usize + 1) as *const u8);

				if (intercept_data.server_time - data.last_event) != 0 {
					println!("testb");
					writeln!(&mut data.outfile,"Delay {}", intercept_data.server_time - data.last_event)
						.expect("Failed to write to outfile.");
					data.last_event = intercept_data.server_time;
				}

				if data.moving {
					writeln!(&mut data.outfile, "MotionNotify {} {}", data.x, data.y)
						.expect("Failed to write to outfile.");
					data.moving = false;
				}

				writeln!(
					&mut data.outfile,
					"{} {}",
					if ev_type == BUTTONPRESS_U8 { "ButtonPress" } else { "ButtonRelease" },
					bc
				).expect("Failed to write to outfile.");
			}
			_ => eprintln!("Unknown event type: {:?}", ev_type)
		}
	}

	data.ev_nr += 1;
	XRecordFreeData(intercept_data)
}

const KEYPRESS_U8: u8 = KeyPress as u8;
const KEYRELEASE_U8: u8 = KeyRelease as u8;
const BUTTONPRESS_U8: u8 = ButtonPress as u8;
const BUTTONRELEASE_U8: u8 = ButtonRelease as u8;
const MOTIONNOTIFY_U8: u8 = MotionNotify as u8;

